const puppeteer = require('puppeteer-core')
const stream = require('stream')
const fs = require('fs')
const path = require('path')
const { exit } = require('process')
const { Stream } = require('stream')

const deviceModel = puppeteer.devices['iPad']

module.exports = NewInstance

// Exposed pagex class constructor.
function NewInstance(options) {
  // Default options.
  this.options = {
    debug: false,
    downloadBrowser: false,
    executePath: '/usr/bin/chromium',
    browserArgs: ['--no-sandbox'],
    logTo: process.stdout,
  }

  Object.assign(this.options, options)

  const opts = this.options

  if (process.env.NODE_ENV === 'test') {
    opts.downloadBrowser = true
  }

  return this
}

// Open browser and a new page, be prepared for nevigating.
NewInstance.prototype.init = async function () {
  let err

  opts = this.options

  if (opts.downloadBrowser) {
    const browserFetcher = puppeteer.createBrowserFetcher()

    const revisionInfo = this._to(await browserFetcher.download('800071'), "dowload browser")

    opts.executePath = revisionInfo.executablePath
    this._log('downloaded browser in path: ' + opts.executePath)
  }

  this.browser = await this._to(
    puppeteer.launch({
      headless: !opts.debug,
      args: opts.browserArgs,
      executablePath: path.resolve(opts.executePath),
    }),
    'launch puppeteer'
  )


  const browser = this.browser

  this.page = await this._to(browser.newPage(), 'open a new page')
  const page = this.page

  await this._to(page.emulate(deviceModel), 'emulate device')

  await page.exposeFunction('__puppeteerLog__', async (str) => {
    this._log(str)
  })

  await page.exposeFunction('__puppeteerLogRaw__', async (str) => {
    this._logRaw(str)
  })

  await page.exposeFunction('__puppeteerFatal__', async (str) => {
    await this.destroy()

    throw new Error(str)
  })

  // chunk is base64 string.
  await page.exposeFunction('__puppeteerWrite__', async (mimeCodec, chunk) => {
    const arr = String.prototype.split.call(mimeCodec, /\/|;/)
    let ext = arr[1]
    if (!ext) {
      throw new Error('invalid type: ' + mimeCodec)
    }
    ext = String.prototype.toLowerCase.call(ext)
    mimeCodec = String.prototype.replace.call(mimeCodec, /\W/g, '&')
    mimeCodec = String.prototype.slice.call(mimeCodec, 0, 5)

    const dir1 = 'download'
    const dir2 = path.basename(new URL(this.pageUrl).pathname)

    let p = path.resolve(__dirname, dir1, dir2, mimeCodec + '.' + ext)

    fs.mkdirSync(path.dirname(p), { recursive: true })

    let buf = Buffer.from(chunk, 'base64')

    return fs.appendFileSync(p, buf)
  })

  await page.evaluateOnNewDocument(async () => {
    await window.__puppeteerLog__('evaluateOnNewDocument')

    // Used to record sourceBuffers.
    window.__puppeteerTypeMap__ = new Map()

    const addSourceBufferOrigin = MediaSource.prototype.addSourceBuffer
    const endOfStreamOrigin = MediaSource.prototype.endOfStream
    const appendBufferOrigin = SourceBuffer.prototype.appendBuffer
    // const addEventListenerOrigin = EventTarget.prototype.addEventListener;

    function seek() {
      // Place seek operation in next eventloop waiting for updated timeEnd value.
      setTimeout(seekNext, 1)

      // Seek to max buffered time.
      function seekNext() {
        const typeMap = window.__puppeteerTypeMap__
        if (typeMap.size <= 0) {
          window.__puppeteerFatal__('typemap size 0')
        }

        let timeEndMin = Number.MAX_VALUE
        typeMap.forEach((val) => {
          if (timeEndMin > val.timeEnd) {
            timeEndMin = val.timeEnd
          }
        })

        const videoEle = document.querySelector('video')
        if (!videoEle) {
          window.__puppeteerFatal__('videoEle null')
        }

        if (videoEle.currentTime < timeEndMin) {
          videoEle.currentTime = timeEndMin
        }

        // Print download progress in percentage.
        let percent =
          (
            (videoEle.currentTime /
              window.__puppeteerMediaSource__.duration) *
            100
          ).toFixed(2) + '%'
        let str = '\r[pagex] dowloading stats: ' + percent
        window.__puppeteerLogRaw__(str)
      }
    }

    MediaSource.prototype.addSourceBuffer = function () {
      window.__puppeteerLog__('addSourceBuffer')

      const sourceBuf = addSourceBufferOrigin.apply(this, arguments)

      sourceBuf.addEventListener('updateend', seek)

      // Record sourceBuffers in document.
      window.__puppeteerTypeMap__.set(sourceBuf, {
        mimeCodec: arguments[0],
        timeEnd: 0,
      })

      // Record the mediaSource in document.
      if (!window.__puppeteerMediaSource__) {
        window.__puppeteerMediaSource__ = this
      } else if (window.__puppeteerMediaSource__ !== this) {
        window.__puppeteerFatal__("mediasource don't match")
      }

      return sourceBuf
    }

    MediaSource.prototype.endOfStream = function () {
      window.__puppeteerLog__('endOfStream')

      Array.prototype.forEach.call(this.sourceBuffers, (e) => {
        e.removeEventListener('updateend', seek)
      })

      // to do
      // end of Download

      return endOfStreamOrigin.apply(this, arguments)
    }

    SourceBuffer.prototype.appendBuffer = function () {
      window.__puppeteerLog__('appendBuffer')

      // Get mimeCodec string corresponding to this sourceBuffer instance.
      const typeObj = window.__puppeteerTypeMap__.get(this)
      if (!typeObj) {
        window.__puppeteerFatal__('type not in typeMap')
      }
      if (!typeObj.mimeCodec) {
        window.__puppeteerFatal__('empty mimeCodec')
      }

      // Update the buffered time end.
      if (this.buffered.length == 1) {
        const t = this.buffered.end(0)

        typeObj.timeEnd = t
      }

      // Serialize the bytes chunk. Needed for tranmitting data between browser and program.
      const arr = new Uint8Array(arguments[0])
      const len = arr.byteLength
      let b = ''
      for (let i = 0; i < len; i++) {
        b += String.fromCharCode(arr[i])
      }
      const sed = window.btoa(b)
      window.__puppeteerWrite__(typeObj.mimeCodec, sed)

      return appendBufferOrigin.apply(this, arguments)
    }
  })
}

// Launch dowload process.
NewInstance.prototype.get = async function (url) {
  const pageUrl = (this.pageUrl = url)
  const page = this.page

  await this._to(page.goto(pageUrl, { waitUntil: 'networkidle2' }), "page goto: " + pageUrl)

  await this._to(
    page.evaluate(async () => {
      const videoEle = document.querySelector('video')
      if (!videoEle) {
        await window.__puppeteerFatal__('no videoEle')
      }

      if (!String.prototype.startsWith.call(videoEle.src, 'blob')) {
        await window.__puppeteerFatal__(
          'not MSE, you should dowload directly from video.src address'
        )
      }

      await window.__puppeteerLog__('video playing')
      return videoEle.play()
    }),
    "page evaluate")
}

// Close the browser.
NewInstance.prototype.destroy = async function () {
  await this._to(this.browser.close(), 'close browser')
}

// Catch promise and log.
NewInstance.prototype._to = async function (pms, msg) {
  let err, data

  this._log(msg + ' \u2610')
    ;[err, data] = await to(pms)
  if (err) {
    throw new Error(msg + ' error: ' + err)
  }

  this._log(msg + ' \u2611')

  return data
}

NewInstance.prototype._logRaw = function (str) {
  tlog(this.options.logTo, str, true)
}

NewInstance.prototype._log = function (str) {
  tlog(this.options.logTo, str, false)
}

function tlog(out, msg, raw) {
  if (out && out instanceof stream.Writable) {
    const logPrefix = '[pagex]'

    let str

    if (!raw) {
      str = `${logPrefix} ${new Date()}: ${msg}\n`
    } else {
      str = msg
    }

    out.write(str)
  }
}

function to(pms) {
  return pms.then((data) => [null, data]).catch((err) => [err, null])
}
